@@ -199,24 +199,27 @@ int pki_ed25519_key_cmp(const ssh_key k1,
  */
 int pki_ed25519_key_dup(ssh_key new, const ssh_key key)
 {
-    if (key->ed25519_privkey == NULL || key->ed25519_pubkey == NULL) {
+    if (key->ed25519_privkey == NULL && key->ed25519_pubkey == NULL) {
         return SSH_ERROR;
     }
 
-    new->ed25519_privkey = malloc(ED25519_SK_LEN);
-    if (new->ed25519_privkey == NULL) {
-        return SSH_ERROR;
+    if (key->ed25519_privkey != NULL) {
+        new->ed25519_privkey = malloc(ED25519_SK_LEN);
+        if (new->ed25519_privkey == NULL) {
+            return SSH_ERROR;
+        }
+        memcpy(new->ed25519_privkey, key->ed25519_privkey, ED25519_SK_LEN);
     }
 
-    new->ed25519_pubkey = malloc(ED25519_PK_LEN);
-    if (new->ed25519_privkey == NULL || new->ed25519_pubkey == NULL){
-        SAFE_FREE(new->ed25519_privkey);
-        return SSH_ERROR;
+    if (key->ed25519_pubkey != NULL) {
+        new->ed25519_pubkey = malloc(ED25519_PK_LEN);
+        if (new->ed25519_pubkey == NULL) {
+            SAFE_FREE(new->ed25519_privkey);
+            return SSH_ERROR;
+        }
+        memcpy(new->ed25519_pubkey, key->ed25519_pubkey, ED25519_PK_LEN);
     }
 
-    memcpy(new->ed25519_privkey, key->ed25519_privkey, ED25519_SK_LEN);
-    memcpy(new->ed25519_pubkey, key->ed25519_pubkey, ED25519_PK_LEN);
-
     return SSH_OK;
 }
 
