@@ -24,22 +24,36 @@
 #include "torture.h"
 #include <libssh/libssh.h>
 
+#include <errno.h>
+#include <sys/types.h>
+#include <pwd.h>
+
+static int sshd_setup(void **state)
+{
+    torture_setup_sshd_server(state);
+
+    return 0;
+}
+
+static int sshd_teardown(void **state) {
+    torture_teardown_sshd_server(state);
+
+    return 0;
+}
+
 static void setup(void **state)
 {
     ssh_session session;
-    const char *host;
-    const char *user;
-    const char *password;
+    struct passwd *pwd;
 
-    host = getenv("TORTURE_HOST");
-    if (host == NULL) {
-        host = "localhost";
-    }
+    pwd = getpwnam("bob");
+    assert_non_null(pwd);
+    setuid(pwd->pw_uid);
 
-    user = getenv("TORTURE_USER");
-    password = getenv("TORTURE_PASSWORD");
-
-    session = torture_ssh_session(host, NULL, user, password);
+    session = torture_ssh_session(TORTURE_SSH_SERVER,
+                                  NULL,
+                                  TORTURE_SSH_USER_ALICE,
+                                  NULL);
 
     assert_false(session == NULL);
     *state = session;
@@ -99,6 +113,7 @@ static void torture_request_env(void **state)
 
 int torture_run_tests(void) {
     int rc;
+    struct torture_state *s = NULL;
 
     UnitTest tests[] = {
         unit_test_setup_teardown(torture_request_env, setup, teardown),
@@ -107,7 +122,9 @@ int torture_run_tests(void) {
     ssh_init();
 
     torture_filter_tests(tests);
+    sshd_setup((void **)&s);
     rc = run_tests(tests);
+    sshd_teardown((void **)&s);
 
     ssh_finalize();
     return rc;
